"""
Import graph data
Calculate Gaussian similarity
Calculate sigmoid similarity
Fuse similarity
Import node walker sampling information
Import node ST sampling information
Calculate Euclidean distance
"""
import csv
import os
import torch as t
import numpy as np
from math import e
import pandas as pd
from scipy import io
from .HypergraphConstructKNN import Eu_dis
import math

def read_csv(path):
    with open(path, "r", newline="") as csv_file:
        reader = csv.reader(csv_file)
        md_data = []
        md_data += [[float(i) for i in row] for row in reader]
        return t.FloatTensor(md_data)


def read_txt(path):
    with open(path, "r", newline="") as txt_file:
        reader = txt_file.readlines()
        md_data = []
        md_data += [[float(i) for i in row.split()] for row in reader]
        return t.FloatTensor(md_data)


def read_mat(path, name):
    matrix = io.loadmat(path)
    matrix = t.FloatTensor(matrix[name])
    return matrix


def read_md_data(path, validation):
    result = [{} for _ in range(validation)]
    for filename in os.listdir(path):
        data_type = filename[filename.index("_") + 1 : filename.index(".") - 1]
        num = int(filename[filename.index(".") - 1])
        result[num - 1][data_type] = read_csv(os.path.join(path, filename))
    return result


def get_edge_index(matrix):
    edge_index = [[], []]
    for i in range(matrix.size(0)):
        for j in range(matrix.size(1)):
            if matrix[i][j] != 0:
                edge_index[0].append(i)
                edge_index[1].append(j)
    return t.LongTensor(edge_index)


def Gauss_M(adj_matrix, N):
    GM = np.zeros((N, N))
    rm = N * 1.0 / sum(sum(adj_matrix * adj_matrix))
    for i in range(N):
        for j in range(N):
            GM[i][j] = e ** (
                -rm
                * (
                    np.dot(
                        adj_matrix[i, :] - adj_matrix[j, :],
                        adj_matrix[i, :] - adj_matrix[j, :],
                    )
                )
            )
    return GM


def Gauss_D(adj_matrix, M):
    GD = np.zeros((M, M))
    T = adj_matrix.transpose()
    rd = M * 1.0 / sum(sum(T * T))
    for i in range(M):
        for j in range(M):
            GD[i][j] = e ** (-rd * (np.dot(T[i] - T[j], T[i] - T[j])))
    return GD


def SigmoidKernelDisease(DiseaseAndRNABinary):
    DiseasePolynomialKernel = []
    counter = 0
    while counter < len(DiseaseAndRNABinary):
        row = []
        counter1 = 0
        while counter1 < len(DiseaseAndRNABinary):
            sum = 0
            counter2 = 0
            while counter2 < len(DiseaseAndRNABinary[counter]):
                sum = (
                    sum
                    + DiseaseAndRNABinary[counter][counter2]
                    * DiseaseAndRNABinary[counter1][counter2]
                )
                counter2 = counter2 + 1
            sum = sum / (782 / len(DiseaseAndRNABinary))
            sum = (math.exp(sum) - math.exp(-sum)) / (math.exp(sum) + math.exp(-sum))
            row.append(sum)
            counter1 = counter1 + 1
        DiseasePolynomialKernel.append(row)
        counter = counter + 1
    return DiseasePolynomialKernel


def SigmoidKernelRNA(DiseaseAndRNABinary):
    MDiseaseAndRNABinary = np.array(DiseaseAndRNABinary)
    RNAAndDiseaseBinary = MDiseaseAndRNABinary.T
    RNAPolynomialKernel = []
    counter = 0
    while counter < len(RNAAndDiseaseBinary):
        row = []
        counter1 = 0
        while counter1 < len(RNAAndDiseaseBinary):
            sum = 0
            counter2 = 0
            while counter2 < len(RNAAndDiseaseBinary[counter]):
                sum = (
                    sum
                    + RNAAndDiseaseBinary[counter][counter2]
                    * RNAAndDiseaseBinary[counter1][counter2]
                )
                counter2 = counter2 + 1
            sum = sum / (782 / len(RNAAndDiseaseBinary))
            sum = (math.exp(sum) - math.exp(-sum)) / (math.exp(sum) + math.exp(-sum))
            row.append(sum)
            counter1 = counter1 + 1
        RNAPolynomialKernel.append(row)
        counter = counter + 1
    return RNAPolynomialKernel


def ReadMyCsv(SaveList, fileName):
    csv_reader = csv.reader(open(fileName))
    for row in csv_reader:
        SaveList.append(row)
    return


def storFile(data, fileName):
    with open(fileName, "w", newline="") as csvfile:
        writer = csv.writer(csvfile)
        writer.writerows(data)
    return


def association_to_matrix(fileName):
    print("Run MMGN")

    df = pd.read_csv(fileName, header=None, names=["node1", "node2"])

    nodes_col = sorted(list(set(df["node1"])))
    nodes_row = sorted(list(set(df["node2"])))

    num_rows = len(nodes_row)
    num_cols = len(nodes_col)
    adj_matrix = [[0] * num_cols for _ in range(num_rows)]

    for _, row in df.iterrows():
        node1_idx = nodes_col.index(row["node1"])
        node2_idx = nodes_row.index(row["node2"])
        adj_matrix[node2_idx][node1_idx] = 1

    adj_df = pd.DataFrame(adj_matrix, columns=nodes_col, index=nodes_row)

    return adj_df


def PrepareBinData(Bin_data_path,SimWalk_RNA_path,SimWalk_Dis_path,SimST_RNA_path,SimST_Dis_path):
    dataset = {}
    mi_dis_data = association_to_matrix(Bin_data_path)

    row_indexes = mi_dis_data.index

    column_indexes = mi_dis_data.columns
    dataset["row_indexes_MDA"] = row_indexes
    dataset["column_indexes_MDA"] = column_indexes

    dataset["md_p"] = t.FloatTensor(np.array(mi_dis_data).T)
    dataset["md_true"] = dataset["md_p"]

    all_zero_index = []
    all_one_index = []
    for i in range(dataset["md_p"].size(0)):
        for j in range(dataset["md_p"].size(1)):
            if dataset["md_p"][i][j] < 1:
                all_zero_index.append([i, j])
            if dataset["md_p"][i][j] >= 1:
                all_one_index.append([i, j])

    np.random.seed(0)
    np.random.shuffle(all_zero_index)
    np.random.shuffle(all_one_index)

    dataset["md"] = []
    zero_tensor = t.LongTensor(all_zero_index)
    one_tensor = t.LongTensor(all_one_index)

    train_zero_index = zero_tensor
    train_one_index = one_tensor

    dataset["md"].append(
        {
            "test": [train_one_index, train_zero_index],
            "train": [train_one_index, train_zero_index],
        }
    )

    DGSM = Gauss_D(dataset["md_p"].numpy(), dataset["md_p"].size(1))
    MGSM = Gauss_M(dataset["md_p"].numpy(), dataset["md_p"].size(0))

    Sigmoid_matrix1 = SigmoidKernelDisease(dataset["md_p"].numpy())
    Sigmoid_matrix2 = SigmoidKernelRNA(dataset["md_p"].numpy())

    Sigmoid_matrix1 = np.array(Sigmoid_matrix1)
    Sigmoid_matrix2 = np.array(Sigmoid_matrix2)

    nd = mi_dis_data.shape[0]
    nm = mi_dis_data.shape[1]
    ID = np.zeros([nd, nd])

    for h1 in range(nd):
        for h2 in range(nd):
            if Sigmoid_matrix2[h1, h2] == 0:
                ID[h1, h2] = DGSM[h1, h2]
            else:
                ID[h1, h2] = (Sigmoid_matrix2[h1, h2] + DGSM[h1, h2]) / 2

    IM = np.zeros([nm, nm])

    for q1 in range(nm):
        for q2 in range(nm):
            if Sigmoid_matrix1[q1, q2] == 0:
                IM[q1, q2] = MGSM[q1, q2]
            else:
                IM[q1, q2] = (Sigmoid_matrix1[q1, q2] + MGSM[q1, q2]) / 2

    dataset["ID"] = t.from_numpy(ID)
    dataset["IM"] = t.from_numpy(IM)

    # walker sampling features
    Walker_Mi = pd.read_csv(SimWalk_RNA_path, header=None, index_col=None)
    Walker_Dis = pd.read_csv(SimWalk_Dis_path, header=None, index_col=None)
    # Calculate the walker Euclidean distance
    Walker_Mi_Eu_dis = np.array(Eu_dis(Walker_Mi))
    Walker_Dis_Eu_dis = np.array(Eu_dis(Walker_Dis))

    dataset["MDA_SD_OD"] = t.from_numpy(Walker_Dis_Eu_dis)
    dataset["MDA_SM_OD"] = t.from_numpy(Walker_Mi_Eu_dis)

    dataset["MDA_SD_S"] = t.from_numpy(np.array(Walker_Dis))
    dataset["MDA_SM_S"] = t.from_numpy(np.array(Walker_Mi))

    # ST sampling features
    Walker_Mi_ST = pd.read_csv(SimST_RNA_path, header=None, index_col=None)
    Walker_Dis_ST = pd.read_csv(SimST_Dis_path, header=None, index_col=None)
    # Calculate the ST Euclidean distance
    Walker_Mi_Eu_dis_ST = np.array(Eu_dis(Walker_Mi_ST))
    Walker_Dis_Eu_dis_ST = np.array(Eu_dis(Walker_Dis_ST))

    dataset["MDA_SD_OD_ST"] = t.from_numpy(Walker_Dis_Eu_dis_ST)
    dataset["MDA_SM_OD_ST"] = t.from_numpy(Walker_Mi_Eu_dis_ST)

    dataset["MDA_SD_S_ST"] = t.from_numpy(np.array(Walker_Dis_ST))
    dataset["MDA_SM_S_ST"] = t.from_numpy(np.array(Walker_Mi_ST))

    return dataset



from __future__ import division

class Dataset_MDA(object):
    def __init__(self, validation, dataset):
        self.data_set = dataset
        self.nums = validation

    def __getitem__(self, index):

        # ID IM: similarity fusion matrix
        # md train test: Relationship between positive and negative samples
        # md_p md_true: Adjacency Matrix
        # SD_OD SM_OD: Euclidean distance calculation for random walks
        # SD_S SM_S: Random walk feature
        # SD_S_ST SM_S_ST: Functional similarity features
        # SD_OD_ST SM_OD_ST: Functional similarity Euclidean distance

        return (
            self.data_set["ID"],
            self.data_set["IM"],
            self.data_set["md"][index]["train"],
            self.data_set["md"][index]["test"],
            self.data_set["md_p"],
            self.data_set["md_true"],
            self.data_set["row_indexes_MDA"],
            self.data_set["column_indexes_MDA"],
            self.data_set["MDA_SD_OD"],
            self.data_set["MDA_SM_OD"],
            self.data_set["MDA_SD_S"],
            self.data_set["MDA_SM_S"],
            self.data_set["MDA_SD_OD_ST"],
            self.data_set["MDA_SM_OD_ST"],
            self.data_set["MDA_SD_S_ST"],
            self.data_set["MDA_SM_S_ST"],
        )

    def __len__(self):
        return self.nums



import numpy as np
import torch


def Eu_dis(x):

    x = np.mat(x)

    aa = np.sum(np.multiply(x, x), 1)
    ab = x * x.T
    dist_mat = aa + aa.T - 2 * ab
    dist_mat[dist_mat < 0] = 0

    dist_mat = np.sqrt(dist_mat)
    dist_mat = np.maximum(dist_mat, dist_mat.T)
    return dist_mat


def feature_concat(*F_list, normal_col=False):

    features = None
    for f in F_list:
        if f is not None and f != []:

            if len(f.shape) > 2:
                f = f.reshape(-1, f.shape[-1])

            if normal_col:
                f_max = np.max(np.abs(f), axis=0)
                f = f / f_max

            if features is None:
                features = f
            else:
                features = np.hstack((features, f))
    if normal_col:
        features_max = np.max(np.abs(features), axis=0)
        features = features / features_max
    return features


def hyperedge_concat(*H_list):

    H = None
    for h in H_list:
        # if h is not None and h != []:
        if h is not None and len(h) != 0:
            # for the first H appended to fused hypergraph incidence matrix
            if H is None:
                H = h
            else:
                if type(h) != list:
                    H = np.hstack((H, h))
                else:
                    tmp = []
                    for a, b in zip(H, h):
                        tmp.append(np.hstack((a, b)))
                    H = tmp
    return H


def generate_G_from_H(H, variable_weight=False):

    if type(H) != list:
        return _generate_G_from_H(H, variable_weight)
    else:
        G = []
        for sub_H in H:
            G.append(generate_G_from_H(sub_H, variable_weight))
        return G


def _generate_G_from_H(H, variable_weight=False):

    H = np.array(H)

    n_edge = H.shape[1]
    # the weight of the hyperedge
    W = np.ones(n_edge)

    DV = np.sum(H * W, axis=1)
    DE = np.sum(H, axis=0)

    invDE = np.mat(np.diag(np.power(DE, -1)))
    DV2 = np.mat(np.diag(np.power(DV, -0.5)))

    W = np.mat(np.diag(W))
    H = np.mat(H)
    HT = H.T

    if variable_weight:
        DV2_H = DV2 * H
        invDE_HT_DV2 = invDE * HT * DV2
        return DV2_H, W, invDE_HT_DV2
    else:
        G = DV2 * H * W * invDE * HT * DV2
        G = torch.Tensor(G)
        return G


def construct_H_with_KNN_from_distance(dis_mat, k_neig, is_probH=False, m_prob=1):

    n_obj = dis_mat.shape[0]
    n_edge = n_obj
    H = np.zeros((n_obj, n_edge))
    for center_idx in range(n_obj):
        dis_mat[center_idx, center_idx] = 0
        dis_vec = dis_mat[center_idx]
        nearest_idx = np.array(np.argsort(dis_vec)).squeeze()
        avg_dis = np.average(dis_vec)
        if not np.any(nearest_idx[:k_neig] == center_idx):
            nearest_idx[k_neig - 1] = center_idx

        for node_idx in nearest_idx[:k_neig]:
            if is_probH:
                H[node_idx, center_idx] = np.exp(
                    -dis_vec[0, node_idx] ** 2 / (m_prob * avg_dis) ** 2
                )  ## affinity matrix的计算公式
            else:
                H[node_idx, center_idx] = 1.0
    return H


def construct_H_with_KNN(X, K_neigs, split_diff_scale=False, is_probH=False, m_prob=1):

    if len(X.shape) != 2:
        X = X.reshape(-1, X.shape[-1])

    if type(K_neigs) == int:
        K_neigs = [K_neigs]

    dis_mat = Eu_dis(X)
    H = []
    for k_neig in K_neigs:
        H_tmp = construct_H_with_KNN_from_distance(dis_mat, k_neig, is_probH, m_prob)

        if not split_diff_scale:
            H = hyperedge_concat(H, H_tmp)
        else:
            H.append(H_tmp)

    return H



from torch import nn
from .param import parameter_parser

args = parameter_parser()


class Myloss(nn.Module):
    def __init__(self):
        super(Myloss, self).__init__()

    def forward(self, one_index, zero_index, input, target):
        loss = nn.MSELoss(reduction="none")
        loss_sum = loss(input, target)

        return (1 - args.alpha) * loss_sum[one_index].sum() + args.alpha * loss_sum[
            zero_index
        ].sum()


def get_L2reg(parameters):
    reg = 0
    for param in parameters:
        reg += 0.5 * (param**2).sum()
    return reg



import numpy as np
import torch
from sklearn.cluster import KMeans


def hyperedge_concat(*H_list):

    H = None
    for h in H_list:
        # if h is not None and h != []:
        if h is not None and len(h) != 0:
            # for the first H appended to fused hypergraph incidence matrix
            if H is None:
                H = h
            else:
                if type(h) != list:
                    H = np.hstack((H, h))
                else:
                    tmp = []
                    for a, b in zip(H, h):
                        tmp.append(np.hstack((a, b)))
                    H = tmp
    return H


def _construct_edge_list_from_cluster(X, clusters):
    """
    construct edge list (numpy array) from cluster for single modality
    :param X: feature
    :param clusters: number of clusters for k-means
    :param adjacent_clusters: a node's adjacent clusters
    :param k_neighbors: number of a node's neighbors
    :return:
    """
    N = X.shape[0]
    kmeans = KMeans(n_clusters=clusters, init="k-means++", random_state=0).fit(X)

    assignment = kmeans.labels_

    H = np.zeros([N, clusters])

    for i in range(N):
        H[i, assignment[i]] = 1

    return H


def construct_H_with_Kmeans(X, clusters, split_diff_scale=False):

    if len(X.shape) != 2:
        X = X.reshape(-1, X.shape[-1])

    if type(clusters) == int:
        clusters = [clusters]

    H = []
    for clusters in clusters:
        H_tmp = _construct_edge_list_from_cluster(X, clusters)

        if not split_diff_scale:
            H = hyperedge_concat(H, H_tmp)
        else:
            H.append(H_tmp)
    return H


def _generate_G_from_H(H, variable_weight=False):

    H = np.array(H)

    n_edge = H.shape[1]
    # the weight of the hyperedge
    W = np.ones(n_edge)

    DV = np.sum(H * W, axis=1)
    DE = np.sum(H, axis=0)

    invDE = np.mat(np.diag(np.power(DE, -1)))
    DV2 = np.mat(np.diag(np.power(DV, -0.5)))

    W = np.mat(np.diag(W))
    H = np.mat(H)
    HT = H.T

    if variable_weight:
        DV2_H = DV2 * H
        invDE_HT_DV2 = invDE * HT * DV2
        return DV2_H, W, invDE_HT_DV2
    else:
        G = DV2 * H * W * invDE * HT * DV2
        G = torch.Tensor(G)
        return G


def generate_G_from_H(H, variable_weight=False):

    if type(H) != list:
        return _generate_G_from_H(H, variable_weight)
    else:
        G = []
        for sub_H in H:
            G.append(generate_G_from_H(sub_H, variable_weight))
        return G



"""MMLMGN runner."""
from .DataProcess import PrepareBinData
from torch import optim
from .Module import MMLMGN
from .DataSet import Dataset_MDA
import warnings
from .trainer import train_epoch_MDA, train_epoch_MDA_ST, train_epoch_MDA_TP
from dataclasses import dataclass

warnings.filterwarnings("ignore")

@dataclass
class InputPaths:
    Bin_data_path: str
    SimWalk_RNA_path: str
    SimWalk_Dis_path: str
    SimST_RNA_path: str
    SimST_Dis_path: str
    """ Paths to the input data.
    Parameters:
        Bin_data_path (str): Path to the RNA and disease association data.
        SimWalk_RNA_path (str): Path to the RNA Walker data.
        SimWalk_Dis_path (str): Path to the disease Walker data.
        SimST_RNA_path (str): Path to the RNA ST data.
        SimST_Dis_path (str): Path to the disease ST data.
    """
@dataclass
class OutputPaths:
    Save_RNA_BinFeature_path: str
    Save_DIS_BinFeature_path: str
    Save_RNAWalkerFeature_path: str
    Save_DISWalkerFeature_path: str
    Save_RNA_STFeature_path: str
    Save_DIS_STFeature_path: str
    
    """ Paths to the output data.
        Save_RNA_BinFeature_path (str): Path where RNA binary features are saved.
        Save_DIS_BinFeature_path (str): Path where disease binary features are saved.
        Save_RNAWalkerFeature_path (str): Path where RNA Walker features are saved.
        Save_DISWalkerFeature_path (str): Path where disease Walker features are saved.
        Save_RNA_STFeature_path (str): Path where RNA ST features are saved.
        Save_DIS_STFeature_path (str): Path where disease ST features are saved.
    """
    
def run(input_paths: InputPaths, output_paths: OutputPaths, mi_num=467, dis_num=72, 
        hidden_list=[256, 256], proj_hidden=64, validation=5, epochs=5, lr=0.01):
    """
    Executes the MMLMGN model training and saves the embedding results.

    Parameters:
        input_paths (InputPaths): Container for all input data file paths.
        output_paths (OutputPaths): Container for all output file paths.
        mi_num (int): Number of RNAs. Default is 467.
        dis_num (int): Number of diseases. Default is 72.
        hidden_list (list of int): List of sizes for hidden layers. Default is [256, 256].
        proj_hidden (int): Size of the projection hidden layer. Default is 64.
        validation (int): Number of validations. Default is 5.
        epochs (int): Number of training epochs. Default is 5.
        lr (float): Learning rate for the optimizer. Default is 0.01.
    Returns:
        None
    """
    # Example of how you might use the paths in your function
    dataset = PrepareBinData(
        input_paths.Bin_data_path,
        input_paths.SimWalk_RNA_path,
        input_paths.SimWalk_Dis_path,
        input_paths.SimST_RNA_path,
        input_paths.SimST_Dis_path
    )
    
    model_MDA = MMLMGN(mi_num, dis_num, hidden_list, proj_hidden)
    optimizer = optim.Adam(model_MDA.parameters(), lr=lr)
    train_data_MDA = Dataset_MDA(validation, dataset)

    # Example training function calls using the paths from output_paths
    train_epoch_MDA(model_MDA, train_data_MDA[0], optimizer, epochs, output_paths.Save_RNA_BinFeature_path, output_paths.Save_DIS_BinFeature_path)
    train_epoch_MDA_ST(model_MDA, train_data_MDA[0], optimizer, epochs, output_paths.Save_RNAWalkerFeature_path, output_paths.Save_DISWalkerFeature_path)
    train_epoch_MDA_TP(model_MDA, train_data_MDA[0], optimizer, epochs, output_paths.Save_RNA_STFeature_path, output_paths.Save_DIS_STFeature_path)
    print("Finish")

if __name__ == "__main__":

    run()



"""
MMLMGN model
Multimodal similarity hypergraph contrastive learning
"""
import random
import os
import numpy as np
import torch
import math
from torch import nn
import torch.nn.functional as F
from torch.nn.parameter import Parameter

device = torch.device("cpu")

def seed_torch(seed):
    random.seed(seed)
    os.environ["PYTHONHASHSEED"] = str(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)

seed_torch(seed=1234)

class HGNN_conv(nn.Module):
    """
    Hypergraph Convolution
    """
    def __init__(self, in_ft, out_ft, bias=True):
        super(HGNN_conv, self).__init__()
        self.in_features = in_ft
        self.out_features = out_ft
        self.weight = Parameter(torch.Tensor(in_ft, out_ft))
        if bias:
            self.bias = Parameter(torch.Tensor(out_ft))
        else:
            self.register_parameter("bias", None)
        self.reset_parameters()

    def reset_parameters(self):
        stdv = 1.0 / math.sqrt(self.weight.size(1))
        self.weight.data.uniform_(-stdv, stdv)
        if self.bias is not None:
            self.bias.data.uniform_(-stdv, stdv)

    def forward(self, x, G):
        x = x.matmul(self.weight)
        if self.bias is not None:
            x = x + self.bias
        x = G.matmul(x)
        return x


class HGCN(nn.Module):
    def __init__(self, in_dim, hidden_list, dropout=0.5):
        super(HGCN, self).__init__()
        self.dropout = dropout

        self.hgnn1 = HGNN_conv(in_dim, hidden_list[0])

    def forward(self, x, G):

        x_embed = self.hgnn1(x, G)
        x_embed_1 = F.leaky_relu(x_embed, 0.25)

        return x_embed_1


class CL_HGCN(nn.Module):
    def __init__(self, in_size, hid_list, num_proj_hidden, alpha=0.5):
        super(CL_HGCN, self).__init__()
        self.hgcn1 = HGCN(in_size, hid_list)
        self.hgcn2 = HGCN(in_size, hid_list)

        self.fc1 = torch.nn.Linear(hid_list[-1], num_proj_hidden)
        self.fc2 = torch.nn.Linear(num_proj_hidden, hid_list[-1])

        self.tau = 0.5
        self.alpha = alpha

    def forward(self, x1, adj1, x2, adj2):

        z1 = self.hgcn1(x1, adj1)
        h1 = self.projection(z1)

        z2 = self.hgcn2(x2, adj2)
        h2 = self.projection(z2)

        loss = self.alpha * self.sim(h1, h2) + (1 - self.alpha) * self.sim(h2, h1)

        return z1, z2, loss

    def projection(self, z):
        z = F.elu(self.fc1(z))
        return self.fc2(z)

    def norm_sim(self, z1, z2):
        z1 = F.normalize(z1)
        z2 = F.normalize(z2)
        return torch.mm(z1, z2.t())

    def sim(self, z1, z2):
        f = lambda x: torch.exp(x / self.tau)
        refl_sim = f(self.norm_sim(z1, z1))
        between_sim = f(self.norm_sim(z1, z2))
        loss = -torch.log(
            between_sim.diag()
            / (refl_sim.sum(1) + between_sim.sum(1) - refl_sim.diag())
        )
        loss = loss.sum(dim=-1).mean()
        return loss


class MMLMGN(nn.Module):
    def __init__(self, mi_num, dis_num, hidd_list, num_proj_hidden):
        super(MMLMGN, self).__init__()

        self.CL_HGCN_mi = CL_HGCN(mi_num + dis_num, hidd_list, num_proj_hidden)
        self.CL_HGCN_dis = CL_HGCN(dis_num + mi_num, hidd_list, num_proj_hidden)

        self.linear_x_1 = nn.Linear(512, 256)
        self.linear_x_2 = nn.Linear(256, 128)
        self.linear_x_3 = nn.Linear(128, 64)

        self.linear_y_1 = nn.Linear(512, 256)
        self.linear_y_2 = nn.Linear(256, 128)
        self.linear_y_3 = nn.Linear(128, 64)

    def forward(
        self, concat_mi_tensor, concat_dis_tensor, G_mi_Kn, G_mi_Km, G_dis_Kn, G_dis_Km
    ):

        mi_embedded = concat_mi_tensor
        dis_embedded = concat_dis_tensor

        mi_feature1, mi_feature2, mi_cl_loss = self.CL_HGCN_mi(
            mi_embedded, G_mi_Kn, mi_embedded, G_mi_Km
        )
        mi_concat_feature = torch.cat([mi_feature1, mi_feature2], dim=1)

        dis_feature1, dis_feature2, dis_cl_loss = self.CL_HGCN_dis(
            dis_embedded, G_dis_Kn, dis_embedded, G_dis_Km
        )
        dis_concat_feature = torch.cat([dis_feature1, dis_feature2], dim=1)

        x1 = torch.relu(self.linear_x_1(mi_concat_feature))
        x2 = torch.relu(self.linear_x_2(x1))
        x = torch.relu(self.linear_x_3(x2))

        y1 = torch.relu(self.linear_y_1(dis_concat_feature))
        y2 = torch.relu(self.linear_y_2(y1))
        y = torch.relu(self.linear_y_3(y2))

        score = x.mm(y.t())

        return score, mi_cl_loss, dis_cl_loss, x, y



import torch
import numpy as np
from . import ConstructHW
import warnings
from .utils import get_L2reg, Myloss
import pandas as pd
warnings.filterwarnings("ignore")
device = torch.device("cpu")


def test(
    model,
    data,
    concat_mi_tensor,
    concat_dis_tensor,
    G_mi_Kn,
    G_mi_Km,
    G_dis_Kn,
    G_dis_Km,
):
    model.eval()
    score, _, _, _, _ = model(
        concat_mi_tensor, concat_dis_tensor, G_mi_Kn, G_mi_Km, G_dis_Kn, G_dis_Km
    )
    test_one_index = data[3][0]
    test_zero_index = data[3][1]
    true_one = data[5][test_one_index]
    true_zero = data[5][test_zero_index]

    pre_one = score[test_one_index]
    pre_zero = score[test_zero_index]

    return true_one, true_zero, pre_one, pre_zero


def train_epoch_MDA(model, train_data, optim, epochs, Save_RNA_BinFeature_path, Save_DIS_BinFeature_path):

    model.train()
    regression_crit = Myloss()

    one_index = train_data[2][0]
    zero_index = train_data[2][1]


    dis_sim_integrate_tensor = train_data[0].to(device)
    mi_sim_integrate_tensor = train_data[1].to(device)


    concat_miRNA = np.hstack(
        [train_data[4].numpy(), mi_sim_integrate_tensor.detach().cpu().numpy()]
    )
    concat_mi_tensor = torch.FloatTensor(concat_miRNA)
    concat_mi_tensor = concat_mi_tensor.to(device)

    G_mi_Kn = ConstructHW.constructHW_knn(
        concat_mi_tensor.detach().cpu().numpy(), K_neigs=[13], is_probH=False
    )
    G_mi_Km = ConstructHW.constructHW_kmean(
        concat_mi_tensor.detach().cpu().numpy(), clusters=[9]
    )
    G_mi_Kn = G_mi_Kn.to(device)
    G_mi_Km = G_mi_Km.to(device)

    concat_dis = np.hstack(
        [train_data[4].numpy().T, dis_sim_integrate_tensor.detach().cpu().numpy()]
    )
    concat_dis_tensor = torch.FloatTensor(concat_dis)
    concat_dis_tensor = concat_dis_tensor.to(device)

    G_dis_Kn = ConstructHW.constructHW_knn(
        concat_dis_tensor.detach().cpu().numpy(), K_neigs=[13], is_probH=False
    )
    G_dis_Km = ConstructHW.constructHW_kmean(
        concat_dis_tensor.detach().cpu().numpy(), clusters=[9]
    )
    G_dis_Kn = G_dis_Kn.to(device)
    G_dis_Km = G_dis_Km.to(device)

    X1 = []
    Y1 = []

    for epoch in range(1, epochs + 1):

        score, mi_cl_loss, dis_cl_loss, X1, Y1 = model(
            concat_mi_tensor,
            concat_dis_tensor,
            G_mi_Kn,
            G_mi_Km,
            G_dis_Kn,
            G_dis_Km,
        )

        recover_loss = regression_crit(
            one_index, zero_index, train_data[4].to(device), score
        )
        reg_loss = get_L2reg(model.parameters())

        tol_loss = recover_loss + mi_cl_loss + dis_cl_loss + 0.00001 * reg_loss
        optim.zero_grad()
        tol_loss.backward()
        optim.step()

    true_value_one, true_value_zero, pre_value_one, pre_value_zero = test(
        model,
        train_data,
        concat_mi_tensor,
        concat_dis_tensor,
        G_mi_Kn,
        G_mi_Km,
        G_dis_Kn,
        G_dis_Km,
    )

    X1 = X1.detach().numpy()
    Y1 = Y1.detach().numpy()
    X = pd.DataFrame(X1)
    Y = pd.DataFrame(Y1)

    X.to_csv(Save_RNA_BinFeature_path, index=True)
    Y.to_csv(Save_DIS_BinFeature_path, index=True)

    return 0


def train_epoch_MDA_ST(model, train_data, optim, epochs, Save_RNAWalkerFeature_path, Save_DISWalkerFeature_path):

    model.train()
    regression_crit = Myloss()

    one_index = train_data[2][0]
    zero_index = train_data[2][1]


    dis_Walker_integrate_tensor = train_data[14].to(device)
    mi_Walker_integrate_tensor = train_data[15].to(device)

    dis_Walker_OD_tensor = train_data[12].to(device)
    mi_Walker_OD_tensor = train_data[13].to(device)

    mi_Walker_integrate_tensor1 = np.hstack(
        [train_data[4].numpy(), mi_Walker_integrate_tensor.detach().cpu().numpy()]
    )
    Walker_mi_tensor = torch.FloatTensor(mi_Walker_integrate_tensor1)
    Walker_mi_tensor = Walker_mi_tensor.to(device)

    G_Walker_mi_Kn = ConstructHW.constructHW_knn(
        Walker_mi_tensor.detach().cpu().numpy(), K_neigs=[13], is_probH=False
    )
    G_Walker_mi_Km = ConstructHW.constructHW_kmean(
        Walker_mi_tensor.detach().cpu().numpy(), clusters=[9]
    )
    G_Walker_mi_Kn = G_Walker_mi_Kn.to(device)
    G_Walker_mi_Km = G_Walker_mi_Km.to(device)

    dis_Walker_integrate_tensor1 = np.hstack(
        [train_data[4].numpy().T, dis_Walker_integrate_tensor.detach().cpu().numpy()]
    )
    Walker_dis_tensor = torch.FloatTensor(dis_Walker_integrate_tensor1)
    Walker_dis_tensor = Walker_dis_tensor.to(device)

    G_Walker_dis_Kn = ConstructHW.constructHW_knn(
        Walker_dis_tensor.detach().cpu().numpy(), K_neigs=[13], is_probH=False
    )
    G_Walker_dis_Km = ConstructHW.constructHW_kmean(
        Walker_dis_tensor.detach().cpu().numpy(), clusters=[9]
    )
    G_Walker_dis_Kn = G_Walker_dis_Kn.to(device)
    G_Walker_dis_Km = G_Walker_dis_Km.to(device)

    mi_Walker_OD_tensor1 = np.hstack(
        [train_data[4].numpy(), mi_Walker_OD_tensor.detach().cpu().numpy()]
    )
    Walker_mi_OD_tensor = torch.FloatTensor(mi_Walker_OD_tensor1)
    Walker_mi_OD_tensor = Walker_mi_OD_tensor.to(device)

    dis_Walker_OD_tensor1 = np.hstack(
        [train_data[4].numpy().T, dis_Walker_OD_tensor.detach().cpu().numpy()]
    )
    Walker_dis_OD_tensor = torch.FloatTensor(dis_Walker_OD_tensor1)
    Walker_dis_OD_tensor = Walker_dis_OD_tensor.to(device)

    X1 = []
    Y1 = []

    for epoch in range(1, epochs + 1):

        score_TP, mi_TP_loss, dis_TP_loss, X1, Y1 = model(
            Walker_mi_OD_tensor,
            Walker_dis_OD_tensor,
            G_Walker_mi_Kn,
            G_Walker_mi_Km,
            G_Walker_dis_Kn,
            G_Walker_dis_Km,
        )

        recover_loss_TP = regression_crit(
            one_index, zero_index, train_data[4].to(device), score_TP
        )
        reg_loss_TP = get_L2reg(model.parameters())

        tol_loss_TP = recover_loss_TP + mi_TP_loss + dis_TP_loss + 0.00001 * reg_loss_TP
        optim.zero_grad()
        tol_loss_TP.backward()
        optim.step()

    true_value_one_TP, true_value_zero_TP, pre_value_one_TP, pre_value_zero_TP = test(
        model,
        train_data,
        Walker_mi_OD_tensor,
        Walker_dis_OD_tensor,
        G_Walker_mi_Kn,
        G_Walker_mi_Km,
        G_Walker_dis_Kn,
        G_Walker_dis_Km,
    )

    X1 = X1.detach().numpy()
    Y1 = Y1.detach().numpy()
    X = pd.DataFrame(X1)
    Y = pd.DataFrame(Y1)

    X.to_csv(Save_RNAWalkerFeature_path, index=True)
    Y.to_csv(Save_DISWalkerFeature_path, index=True)

    return 0


def train_epoch_MDA_TP(model, train_data, optim, epochs, Save_RNA_STFeature_path, Save_DIS_STFeature_path):

    model.train()
    regression_crit = Myloss()

    one_index = train_data[2][0]
    zero_index = train_data[2][1]

    dis_Walker_integrate_tensor = train_data[10].to(device)
    mi_Walker_integrate_tensor = train_data[11].to(device)

    dis_Walker_OD_tensor = train_data[8].to(device)
    mi_Walker_OD_tensor = train_data[9].to(device)

    mi_Walker_integrate_tensor1 = np.hstack(
        [train_data[4].numpy(), mi_Walker_integrate_tensor.detach().cpu().numpy()]
    )
    Walker_mi_tensor = torch.FloatTensor(mi_Walker_integrate_tensor1)
    Walker_mi_tensor = Walker_mi_tensor.to(device)

    G_Walker_mi_Kn = ConstructHW.constructHW_knn(
        Walker_mi_tensor.detach().cpu().numpy(), K_neigs=[13], is_probH=False
    )
    G_Walker_mi_Km = ConstructHW.constructHW_kmean(
        Walker_mi_tensor.detach().cpu().numpy(), clusters=[9]
    )
    G_Walker_mi_Kn = G_Walker_mi_Kn.to(device)
    G_Walker_mi_Km = G_Walker_mi_Km.to(device)


    dis_Walker_integrate_tensor1 = np.hstack(
        [train_data[4].numpy().T, dis_Walker_integrate_tensor.detach().cpu().numpy()]
    )
    Walker_dis_tensor = torch.FloatTensor(dis_Walker_integrate_tensor1)
    Walker_dis_tensor = Walker_dis_tensor.to(device)

    G_Walker_dis_Kn = ConstructHW.constructHW_knn(
        Walker_dis_tensor.detach().cpu().numpy(), K_neigs=[13], is_probH=False
    )
    G_Walker_dis_Km = ConstructHW.constructHW_kmean(
        Walker_dis_tensor.detach().cpu().numpy(), clusters=[9]
    )
    G_Walker_dis_Kn = G_Walker_dis_Kn.to(device)
    G_Walker_dis_Km = G_Walker_dis_Km.to(device)


    mi_Walker_OD_tensor1 = np.hstack(
        [train_data[4].numpy(), mi_Walker_OD_tensor.detach().cpu().numpy()]
    )
    Walker_mi_OD_tensor = torch.FloatTensor(mi_Walker_OD_tensor1)
    Walker_mi_OD_tensor = Walker_mi_OD_tensor.to(device)

    dis_Walker_OD_tensor1 = np.hstack(
        [train_data[4].numpy().T, dis_Walker_OD_tensor.detach().cpu().numpy()]
    )
    Walker_dis_OD_tensor = torch.FloatTensor(dis_Walker_OD_tensor1)
    Walker_dis_OD_tensor = Walker_dis_OD_tensor.to(device)

    X1 = []
    Y1 = []

    for epoch in range(1, epochs + 1):

        score_TP, mi_TP_loss, dis_TP_loss, X1, Y1 = model(
            Walker_mi_OD_tensor,
            Walker_dis_OD_tensor,
            G_Walker_mi_Kn,
            G_Walker_mi_Km,
            G_Walker_dis_Kn,
            G_Walker_dis_Km,
        )

        recover_loss_TP = regression_crit(
            one_index, zero_index, train_data[4].to(device), score_TP
        )
        reg_loss_TP = get_L2reg(model.parameters())

        tol_loss_TP = recover_loss_TP + mi_TP_loss + dis_TP_loss + 0.00001 * reg_loss_TP
        optim.zero_grad()
        tol_loss_TP.backward()
        optim.step()

    true_value_one_TP, true_value_zero_TP, pre_value_one_TP, pre_value_zero_TP = test(
        model,
        train_data,
        Walker_mi_OD_tensor,
        Walker_dis_OD_tensor,
        G_Walker_mi_Kn,
        G_Walker_mi_Km,
        G_Walker_dis_Kn,
        G_Walker_dis_Km,
    )

    X1 = X1.detach().numpy()
    Y1 = Y1.detach().numpy()
    X = pd.DataFrame(X1)
    Y = pd.DataFrame(Y1)

    X.to_csv(Save_RNA_STFeature_path, index=True)
    Y.to_csv(Save_DIS_STFeature_path, index=True)

    return 0



from . import HypergraphConstructKNN
from . import HypergraphConstructKmeans


def constructHW_knn(X, K_neigs, is_probH):
    """incidence matrix"""
    H = HypergraphConstructKNN.construct_H_with_KNN(X, K_neigs, is_probH)

    G = HypergraphConstructKNN._generate_G_from_H(H)

    return G


def constructHW_kmean(X, clusters):
    """incidence matrix"""
    H = HypergraphConstructKmeans.construct_H_with_Kmeans(X, clusters)

    G = HypergraphConstructKmeans._generate_G_from_H(H)

    return G



import argparse


def parameter_parser():

    parser = argparse.ArgumentParser(description="Run MGNN.")

    parser.add_argument(
        "--Bin_data_path",
        type=str,
        default="../data/NodeGraph.csv",
        help="Path to the association of RNA and disease",
    )

    parser.add_argument(
        "--SimWalk_RNA_path",
        type=str,
        default="../data/Node_A_Walker.csv",
        help="Path to the walker of RNA",
    )

    parser.add_argument(
        "--SimWalk_Dis_path",
        type=str,
        default="../data/Node_B_Walker.csv",
        help="Path to the walker of disease",
    )

    parser.add_argument(
        "--SimST_RNA_path",
        type=str,
        default="../data/Node_A_St.csv",
        help="Path to the ST of RNA",
    )

    parser.add_argument(
        "--SimST_Dis_path",
        type=str,
        default="../data/Node_B_St.csv",
        help="Path to the ST of disease",
    )

    parser.add_argument(
        "--validation", type=int, default=2, help="the number of validation"
    )
    parser.add_argument("--epoch", type=int, default=5, help="the number of epoch.")
    parser.add_argument("--mi_num", type=int, default=467, help="the number of RNA")
    parser.add_argument(
        "--dis_num", type=int, default=72, help="the number of diseases"
    )
    parser.add_argument("--alpha", type=int, default=0.11, help="the size of alpha")
    parser.add_argument(
        "--dropout", type=float, default=0.1, help="Dropout rate (1 - keep probability)"
    )
    parser.add_argument(
        "--proj_hidden", type=int, default=64, help="the number of proj_hidden"
    )

    parser.add_argument(
        "--hidden_list", nargs="+", type=int, help="the number of hidden_list"
    )
    parser.set_defaults(hidden_list=[256, 256])

    parser.add_argument(
        "--Save_RNA_BinFeature_path",
        type=str,
        default="../output/NodeAGraphEmb.csv",
        help="Path to save the node A Graph processed file",
    )

    parser.add_argument(
        "--Save_DIS_BinFeature_path",
        type=str,
        default="../output/NodeBGraphEmb.csv",
        help="Path to save the node B Graph processed file",
    )

    parser.add_argument(
        "--Save_RNAWalkerFeature_path",
        type=str,
        default="../output/NodeAWalkerEmb.csv",
        help="Path to save the node A Walker processed file",
    )

    parser.add_argument(
        "--Save_DISWalkerFeature_path",
        type=str,
        default="../output/NodeBWalkerEmb.csv",
        help="Path to save the node B Walker processed file",
    )

    parser.add_argument(
        "--Save_RNA_STFeature_path",
        type=str,
        default="../output/NodeAStEmb.csv",
        help="Path to save the node A ST processed file",
    )

    parser.add_argument(
        "--Save_DIS_STFeature_path",
        type=str,
        default="../output/NodeBStEmb.csv",
        help="Path to save the node B ST processed file",
    )

    return parser.parse_args()



from setuptools import setup, find_packages
import codecs
import os

here = os.path.abspath(os.path.dirname(__file__))

with codecs.open(os.path.join(here, "README.md"), encoding="utf-8") as fh:
    long_description = "\n" + fh.read()

VERSION = "0.1.1"
DESCRIPTION = "Hypergraph contrastive learning framework for multiple similarity modalities"
LONG_DESCRIPTION = "A package that demos mmlmgn model with Mdata."

# Setting up
setup(
    name="mmlmgn",
    version=VERSION,
    author="axin",
    author_email="xinfei106@gmail.com",
    description=DESCRIPTION,
    long_description_content_type="text/markdown",
    long_description=long_description,
    packages=find_packages(),
    url="https://github.com/1axin/MML-MGNN",
    install_requires = [
        "numpy>=1.22.4,<2.0",
        "pandas>=2.2.3",
        "scikit-learn>=1.5.2",
        "scipy>=1.13.1",
        "torch>=1.10.0",
        "joblib>=1.4.2"
    ],
    keywords=["Multiple Similarity Modes", "HGCN", "Hypergraph Contrastive Learning"],
    classifiers=[
        "Development Status :: 1 - Planning",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Operating System :: Unix",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ],
    package_data={"": ["*.csv", "*.txt", ".toml"]},
    include_package_data=True,
)
